import tempfile
from pathlib import Path

import numpy as np

from hyrax import Hyrax


class DatasetRequestBenchmarks:
    """Timing benchmarks for requesting data from the Hyrax random dataset"""

    def setup(self):
        """Prepare for benchmark by defining and setting up a random dataset"""
        self.tmp_dir = tempfile.TemporaryDirectory()
        self.input_dir = Path(self.tmp_dir.name)

        self.h = Hyrax()
        self.h.config["general"]["results_dir"] = str(self.input_dir)
        self.h.config["data_set"]["name"] = "HyraxRandomDataset"

        num_vectors = 4096
        self.h.config["data_set"]["HyraxRandomDataset"]["size"] = num_vectors
        self.h.config["data_set"]["HyraxRandomDataset"]["seed"] = 0
        self.h.config["data_set"]["HyraxRandomDataset"]["shape"] = [3, 64, 64]

        self.ds = self.h.prepare()

        self.indexes = np.random.randint(0, num_vectors, size=128, dtype=int)

    def time_request_all_data(self):
        """Benchmark the amount of time needed to retrieve all the data from
        the random dataset
        """
        for indx in self.indexes:
            self.ds[indx]



import numpy as np
import pytest

import hyrax
from hyrax.data_sets.data_provider import DataProvider
from hyrax.data_sets.inference_dataset import InferenceDataSet, InferenceDataSetWriter


@pytest.fixture(scope="session", params=[1, 2, 3, 4, 5])
def inference_dataset(tmp_path_factory, request):
    """Fixture where I write test data in an InferenceDataSetWriter
    It returns the data written"""
    h = hyrax.Hyrax()
    h.config["general"]["dev_mode"] = True
    h.config["data_set"]["name"] = "HyraxRandomDataset"
    h.config["data_set"]["HyraxRandomDataset"]["size"] = 20
    h.config["data_set"]["HyraxRandomDataset"]["seed"] = 0
    h.config["data_set"]["HyraxRandomDataset"]["shape"] = [2]
    original_data_set = h.prepare()

    current_data_set = original_data_set

    for round_number in range(request.param):
        tmp_path = tmp_path_factory.mktemp(f"order_test_{request.param}_{round_number}")

        data_writer = InferenceDataSetWriter(current_data_set, tmp_path)

        indexes = np.array(range(20))
        np.random.shuffle(indexes)

        data_set_ids = np.array(list(current_data_set.ids()))

        data_writer.write_batch(
            np.array(data_set_ids[indexes[0:10]]),  # ids
            get_data_by_dataset_type(current_data_set, indexes[0:10]),  # Results
        )

        data_writer.write_batch(
            np.array(data_set_ids[indexes[10:20]]),  # ids
            get_data_by_dataset_type(current_data_set, indexes[10:20]),  # Results
        )
        data_writer.write_index()
        current_data_set = InferenceDataSet(h.config, tmp_path)

    return original_data_set, current_data_set


def get_data_by_dataset_type(dataset, idx):
    """Different behavior depending on whether the dataset is an `InferenceDataset`
    vs. a DataProvider dataset. IF it's an InferenceDataset, we return the data
    directly, otherwise we unpack the data from the DataProvider."""
    output_data = dataset[idx]
    if isinstance(dataset, DataProvider):
        output_data = output_data["data"]["image"]

    return np.array(output_data)


def test_order(inference_dataset):
    """Test ID and metadata ordering consistency between original and inference datasets.

    Test cases:
    1) ids() should not be in the same order between original and result
    2) ids() should contain all the IDs in the original dataset
    3) the ids() from ids, and the ids delivered via metadata from the inference dataset should match exactly
    4) The value from inference_dataset[idx] should match a value from data_set
    4a) The matching values should have the same ID in both inference_dataset and data_set according to .ids()
    4b) The matching values should have the same ID in both inference_dataset and data_set according to
        respective metadata
    5) metadata() must preserve the exact order of requested indices (critical for visualization correctness)
    """
    orig, result = inference_dataset

    orig_ids = list(orig.ids())
    result_ids = list(result.ids())

    all_idx = list(range(20))
    orig_meta_ids = list(orig.metadata(all_idx, ["object_id_data"])["object_id_data"])
    result_meta_ids = list(result.metadata(all_idx, ["object_id_data"])["object_id_data"])

    # Check no IDs are dropped
    for id in orig_ids:
        assert id in result_ids
        assert id in orig_meta_ids
        assert id in result_meta_ids

    # Check all data is the correct data for that ID
    for result_i in range(20):
        for orig_i in range(20):
            if np.all(orig[orig_i]["data"]["image"] == result[result_i].numpy()):
                try:
                    assert orig_ids[orig_i] == result_ids[result_i]
                    assert orig_meta_ids[orig_i] == result_meta_ids[result_i]
                    assert orig_ids[orig_i] == result_meta_ids[result_i]
                    assert orig_meta_ids[orig_i] == result_ids[result_i]
                except Exception as e:
                    print(f"Original ID: {orig_ids[orig_i]} (Correct b/c data matches)")
                    print(f"Original metaID: {orig_meta_ids[orig_i]}")
                    print(f"Result ID: {result_ids[result_i]}")
                    print(f"Result metaID: {result_meta_ids[result_i]}")
                    raise e
                break
        else:
            assert False, "Could not find matching value for ID."  # noqa: B011

    # Test explicit metadata ordering preservation with non-sequential patterns
    test_patterns = [
        [3, 1, 4, 0, 2],
        [19, 5, 10, 15, 2],
        [0, 19, 1, 18],
    ]

    for idx_pattern in test_patterns:
        # Skip patterns that exceed dataset size
        if max(idx_pattern) >= len(result):
            continue

        metadata_result = result.metadata(idx_pattern, ["object_id_data"])

        # Get expected IDs in the exact order requested
        expected_ids = [list(result.ids())[i] for i in idx_pattern]
        actual_ids = [str(id) for id in metadata_result["object_id_data"]]

        assert actual_ids == expected_ids, (
            f"CRITICAL: Metadata ordering broken! For indices {idx_pattern}:\n"
            f"Expected IDs: {expected_ids}\n"
            f"Actual IDs:   {actual_ids}\n"
            f"This will cause visualization labels and data to be scrambled."
        )



# ruff: noqa: D101, D102


import torch
import torch.nn as nn
import torch.nn.functional as F  # noqa N812
from torch import Tensor
from torchvision.transforms.v2 import CenterCrop

# extra long import here to address a circular import issue
from hyrax.models.model_registry import hyrax_model


class ArcsinhActivation(nn.Module):
    """Helper module for HyraxAutoencoderV2 to use the arcsinh function"""

    def forward(self, x):
        return torch.arcsinh(x)


@hyrax_model
class HyraxAutoencoderV2(nn.Module):
    """
    This is tweaked version of HyraxAutoencoder and is designed to work with a wide range of imaging datasets.

    V2 improvements:
    - Configurable final layer activation
    - Uses criterion and optimizer from config variables
    """

    def __init__(self, config, shape=(5, 250, 250)):
        super().__init__()
        self.config = config

        # TODO config-ize or get from data loader somehow
        self.num_input_channels, self.image_width, self.image_height = shape

        self.c_hid = self.config["model"]["base_channel_size"]
        self.latent_dim = self.config["model"]["latent_dim"]

        # Calculate how much our convolutional layers will affect the size of final convolution
        # Formula evaluated from: https://pytorch.org/docs/stable/generated/torch.nn.Conv2d.html
        #
        # If the number of layers are changed this will need to be rewritten.
        self.conv_end_w = self.conv2d_multi_layer(self.image_width, 3, kernel_size=3, padding=1, stride=2)
        self.conv_end_h = self.conv2d_multi_layer(self.image_height, 3, kernel_size=3, padding=1, stride=2)

        self._init_encoder()
        self._init_decoder()

    def conv2d_multi_layer(self, input_size, num_applications, **kwargs) -> int:
        for _ in range(num_applications):
            input_size = self.conv2d_output_size(input_size, **kwargs)

        return int(input_size)

    def conv2d_output_size(self, input_size, kernel_size, padding=0, stride=1, dilation=1) -> int:
        # From https://pytorch.org/docs/stable/generated/torch.nn.Conv2d.html
        numerator = input_size + 2 * padding - dilation * (kernel_size - 1) - 1
        return int((numerator / stride) + 1)

    def _init_encoder(self):
        self.encoder = nn.Sequential(
            nn.Conv2d(self.num_input_channels, self.c_hid, kernel_size=3, padding=1, stride=2),
            nn.GELU(),
            nn.Conv2d(self.c_hid, self.c_hid, kernel_size=3, padding=1),
            nn.GELU(),
            nn.Conv2d(self.c_hid, 2 * self.c_hid, kernel_size=3, padding=1, stride=2),
            nn.GELU(),
            nn.Conv2d(2 * self.c_hid, 2 * self.c_hid, kernel_size=3, padding=1),
            nn.GELU(),
            nn.Conv2d(2 * self.c_hid, 2 * self.c_hid, kernel_size=3, padding=1, stride=2),
            nn.GELU(),
            nn.Flatten(),  # Image grid to single feature vector
            nn.Linear(2 * self.conv_end_h * self.conv_end_w * self.c_hid, self.latent_dim),
        )

    def _eval_encoder(self, x):
        return self.encoder(x)

    def _init_decoder(self):
        self.dec_linear = nn.Sequential(
            nn.Linear(self.latent_dim, 2 * self.conv_end_h * self.conv_end_w * self.c_hid), nn.GELU()
        )

        # Configure final activation
        # Should be set to the same value as ["dataset"]["transform"] in most cases
        final_layer = self.config["model"].get("final_layer", "tanh")
        if final_layer == "sigmoid":
            self.final_activation = nn.Sigmoid()
        elif final_layer == "tanh":
            self.final_activation = nn.Tanh()
        elif final_layer == "arcsinh":
            self.final_activation = ArcsinhActivation()
        elif final_layer == "identity":
            self.final_activation = nn.Identity()
        else:
            self.final_activation = nn.Tanh()

        self.decoder = nn.Sequential(
            nn.ConvTranspose2d(
                2 * self.c_hid, 2 * self.c_hid, kernel_size=3, output_padding=1, padding=1, stride=2
            ),  # 4x4 => 8x8
            nn.GELU(),
            nn.Conv2d(2 * self.c_hid, 2 * self.c_hid, kernel_size=3, padding=1),
            nn.GELU(),
            nn.ConvTranspose2d(
                2 * self.c_hid, self.c_hid, kernel_size=3, output_padding=1, padding=1, stride=2
            ),  # 8x8 => 16x16
            nn.GELU(),
            nn.Conv2d(self.c_hid, self.c_hid, kernel_size=3, padding=1),
            nn.GELU(),
            nn.ConvTranspose2d(
                self.c_hid, self.num_input_channels, kernel_size=3, output_padding=1, padding=1, stride=2
            ),  # 16x16 => 32x32
            self.final_activation,
        )

    def _eval_decoder(self, x):
        x = self.dec_linear(x)
        x = x.reshape(x.shape[0], -1, self.conv_end_h, self.conv_end_w)
        x = self.decoder(x)
        x = CenterCrop(size=(self.image_width, self.image_height))(x)
        return x

    def forward(self, batch):
        return self._eval_encoder(batch)

    def train_step(self, batch):
        """This function contains the logic for a single training step. i.e. the
        contents of the inner loop of a ML training process.

        Parameters
        ----------
        batch : tuple
            A tuple containing the inputs and labels for the current batch.

        Returns
        -------
        Current loss value : dict
            Dictionary containing the loss value for the current batch.
        """
        x = batch
        z = self._eval_encoder(x)
        x_hat = self._eval_decoder(z)

        # Configurable band reduction strategy
        band_reduction = self.config["criterion"].get("band_loss_reduction", "mean")

        # The loss averaging strategy here is different from v1 which averages
        # over only the batch dimension. Here we always average over both batch
        # and spaital dimensions; so as the loss-value is not impacted by image size.
        if band_reduction == "sum":
            # Sum across bands, mean over spatial dims and batch
            # More channels will result in larger loss values
            # but MIGHT result in better popping out of bad reconstruction
            # in a single band/channel
            criterion_cls = type(self.criterion)
            loss = criterion_cls(reduction="none")(x_hat, x)
            loss = loss.sum(dim=1).mean()
        elif band_reduction == "mean":
            # Default: Mean over all dimensions (batch,channel,spatial)
            loss = self.criterion(x_hat, x)
        else:
            raise ValueError(
                f"band_loss_reduction:{band_reduction} not supported by HyraxAutoencoderV2.\
                               Current supported options are sum and mean (default)"
            )

        self.optimizer.zero_grad()

        loss.backward()

        self.optimizer.step()

        return {"loss": loss.item()}

    @staticmethod
    def to_tensor(data_dict) -> tuple[Tensor]:
        """This function converts structured data to the input tensor we need to run

        Parameters
        ----------
        data_dict : dict
            The dictionary returned from our data source
        """
        if "image" in data_dict:
            return data_dict["image"]
        else:
            raise RuntimeError("Data dict did not contain image key.")



import logging

from .verb_registry import Verb, hyrax_verb

logger = logging.getLogger(__name__)


@hyrax_verb
class Model(Verb):
    """Resolves the model class that is defined in the config file.
    This will return a reference to the model class."""

    cli_name = "model"
    add_parser_kwargs = {}

    @staticmethod
    def setup_parser():
        """Not implemented"""
        pass

    def run_cli(self):
        """Not implemented"""
        logger.error("Running model from the cli is unimplemented")

    def run(self):
        """Fetch and return the model _class_. Does not create an instance of
        the model class.
        """
        from hyrax.models.model_registry import fetch_model_class

        config = self.config

        model_cls = fetch_model_class(config)

        return model_cls



# ruff: noqa: D101, D102

# This is a more flexible version of hsc_dcae.py that should
# work with a variety of image sizes and includes a true latent bottleneck
# for better anomaly detection capabilities.

import torch
import torch.nn as nn
import torch.nn.functional as f
import torch.optim as optim

# extra long import here to address a circular import issue
from hyrax.models.model_registry import hyrax_model


class ArcsinhActivation(nn.Module):
    """Helper module for ImageDCAE to use the arcsinh function"""

    def forward(self, x):
        return torch.arcsinh(x)


@hyrax_model
class ImageDCAE(nn.Module):
    """
    This is an autoencoder with skipconnections that should work with
    arbitarily sized images with arbitrary number of channels.
    """

    def __init__(self, config, shape):
        super().__init__()

        # Store input shape for dynamic sizing
        self.input_shape = shape
        self.config = config

        # Extract dimensions from input shape
        if len(shape) == 4:  # Batch dimension included
            self.num_input_channels = shape[1]
            self.image_height, self.image_width = shape[2], shape[3]
        else:  # No batch dimension
            self.num_input_channels = shape[0]
            self.image_height, self.image_width = shape[1], shape[2]

        # Get latent dimension from config (similar to HyraxAutoencoder)
        self.latent_dim = config["model"].get("latent_dim", 512)
        self.base_channel_size = config["model"].get("base_channel_size", 32)

        # Calculate the size after convolutional layers for the linear bottleneck
        self.conv_output_size = self._calculate_conv_output_size()

        in_channels = self.num_input_channels

        # Encoder - using configurable base channel size
        self.encoder1 = nn.Conv2d(in_channels, self.base_channel_size, kernel_size=3, stride=1, padding=1)
        self.encoder2 = nn.Conv2d(
            self.base_channel_size, self.base_channel_size * 2, kernel_size=3, stride=1, padding=1
        )
        self.encoder3 = nn.Conv2d(
            self.base_channel_size * 2, self.base_channel_size * 4, kernel_size=3, stride=1, padding=1
        )
        self.encoder4 = nn.Conv2d(
            self.base_channel_size * 4, self.base_channel_size * 8, kernel_size=3, stride=1, padding=1
        )

        self.pool = nn.MaxPool2d(2, 2)

        # Latent bottleneck - similar to HyraxAutoencoder
        self.latent_encoder = nn.Sequential(
            nn.Flatten(), nn.Linear(self.conv_output_size, self.latent_dim), nn.GELU()
        )

        self.latent_decoder = nn.Sequential(nn.Linear(self.latent_dim, self.conv_output_size), nn.GELU())

        # Decoder - using normal Conv2d with upsampling instead of ConvTranspose2d
        # This approach is more flexible for different image sizes
        self.decoder4 = nn.Conv2d(
            self.base_channel_size * 8, self.base_channel_size * 4, kernel_size=3, stride=1, padding=1
        )
        self.decoder3 = nn.Conv2d(
            self.base_channel_size * 4, self.base_channel_size * 2, kernel_size=3, stride=1, padding=1
        )
        self.decoder2 = nn.Conv2d(
            self.base_channel_size * 2, self.base_channel_size, kernel_size=3, stride=1, padding=1
        )
        self.decoder1 = nn.Conv2d(self.base_channel_size, in_channels, kernel_size=3, stride=1, padding=1)

        self.activation = nn.GELU()  # Better gradients than ReLU

        # Configure final activation
        final_layer = config["model"].get("final_layer", "identity")
        if final_layer == "sigmoid":
            self.final_activation = nn.Sigmoid()
        elif final_layer == "tanh":
            self.final_activation = nn.Tanh()
        elif final_layer == "arcsinh":
            self.final_activation = ArcsinhActivation()
        else:
            self.final_activation = nn.Identity()

    def _calculate_conv_output_size(self):
        """Calculate the output size after all convolutional layers for the linear bottleneck."""
        # Simulate the forward pass through conv layers to get the size
        h, w = self.image_height, self.image_width

        # After 3 pooling operations (each divides by 2)
        h = h // 8
        w = w // 8

        # Final feature map size: (base_channel_size * 8) * h * w
        return (self.base_channel_size * 8) * h * w

    def encode(self, x):
        """Encode input to latent space with skip connections."""
        # Encoder with skip connections
        x1 = self.activation(self.encoder1(x))
        p1 = self.pool(x1)

        x2 = self.activation(self.encoder2(p1))
        p2 = self.pool(x2)

        x3 = self.activation(self.encoder3(p2))
        p3 = self.pool(x3)

        x4 = self.activation(self.encoder4(p3))

        # Store the spatial dimensions for reconstruction
        self.encoded_spatial_shape = x4.shape

        # Pass through latent bottleneck
        latent = self.latent_encoder(x4)

        return latent, [x3, x2, x1], x4.shape

    def decode(self, latent, skip_connections, encoded_shape):
        """Decode from latent space to image with skip connections."""
        # Reconstruct from latent space
        x = self.latent_decoder(latent)

        # Reshape back to convolutional feature map
        x = x.reshape(encoded_shape)

        # Decoder with skip connections and dynamic upsampling
        x = f.interpolate(x, size=skip_connections[0].shape[2:], mode="bilinear", align_corners=False)
        x = self.activation(self.decoder4(x) + skip_connections[0])

        x = f.interpolate(x, size=skip_connections[1].shape[2:], mode="bilinear", align_corners=False)
        x = self.activation(self.decoder3(x) + skip_connections[1])

        x = f.interpolate(x, size=skip_connections[2].shape[2:], mode="bilinear", align_corners=False)
        x = self.activation(self.decoder2(x) + skip_connections[2])

        # Final interpolation to input size
        if hasattr(self, "original_size"):
            x = f.interpolate(x, size=self.original_size, mode="bilinear", align_corners=False)

        x = self.final_activation(self.decoder1(x))

        return x

    def forward(self, x):
        """Forward pass - returns latent representation for anomaly detection."""
        # Store original spatial dimensions for decoding
        self.original_size = x.shape[2:]

        # Encode to latent space
        latent, skip_connections, encoded_shape = self.encode(x)

        return latent

    def reconstruct(self, x):
        """Full reconstruction for evaluation and anomaly detection."""
        # Dropping labels if present
        x = x[0] if isinstance(x, tuple) else x

        # Store original spatial dimensions for decoding
        self.original_size = x.shape[2:]

        # Encode to latent space
        latent, skip_connections, encoded_shape = self.encode(x)

        # Decode back to image
        reconstructed = self.decode(latent, skip_connections, encoded_shape)

        return reconstructed

    def train_step(self, batch):
        """This function contains the logic for a single training step.

        Parameters
        ----------
        batch : tuple
            A tuple containing the two values the loss function

        Returns
        -------
        Current loss value : dict
            Dictionary containing the loss value for the current batch.
        """
        data = batch
        self.optimizer.zero_grad()

        # Store original spatial dimensions for decoding
        self.original_size = data.shape[2:]

        # Encode to latent space
        latent, skip_connections, encoded_shape = self.encode(data)

        # Decode back to image
        decoded = self.decode(latent, skip_connections, encoded_shape)

        # Compute loss
        loss = self.criterion(decoded, data)
        loss.backward()
        self.optimizer.step()

        return {"loss": loss.item()}

    def _optimizer(self):
        """Default optimizer configuration."""
        return optim.Adam(self.parameters(), lr=1e-3)

    @staticmethod
    def to_tensor(data_dict):
        """Convert structured data to tensor format."""
        if "image" in data_dict:
            return data_dict["image"]
        else:
            raise RuntimeError("Data dict did not contain image key.")



import logging
from pathlib import Path

import mlflow
from tensorboardX import SummaryWriter

from hyrax.config_utils import create_results_dir, log_runtime_config
from hyrax.gpu_monitor import GpuMonitor
from hyrax.model_exporters import export_to_onnx
from hyrax.pytorch_ignite import (
    create_trainer,
    create_validator,
    dist_data_loader,
    setup_dataset,
    setup_model,
)

from .verb_registry import Verb, hyrax_verb

logger = logging.getLogger(__name__)


@hyrax_verb
class Train(Verb):
    """Train verb"""

    cli_name = "train"
    add_parser_kwargs = {}

    @staticmethod
    def setup_parser(parser):
        """We don't need any parser setup for CLI opts"""
        pass

    def run_cli(self, args=None):
        """CLI stub for Train verb"""
        logger.info("train run from CLI.")

        self.run()

    def run(self):
        """
        Run the training process for the configured model and data loader.
        Returns the trained model.

        """
        import inspect

        config = self.config

        # Create a results directory
        results_dir = create_results_dir(config, "train")
        log_runtime_config(config, results_dir)

        # Create a tensorboardX logger
        tensorboardx_logger = SummaryWriter(log_dir=results_dir)

        # Instantiate the model and dataset
        dataset = setup_dataset(config, tensorboardx_logger)
        model = setup_model(config, dataset)

        # Create a data loader for the training set (and validation split if configured)
        data_loaders = dist_data_loader(dataset, config, ["train", "validate"])
        train_data_loader, _ = data_loaders["train"]
        validation_data_loader, _ = data_loaders.get("validate", (None, None))

        # Create trainer, a pytorch-ignite `Engine` object
        trainer = create_trainer(model, config, results_dir, tensorboardx_logger)

        # Create a validator if a validation data loader is available
        if validation_data_loader is not None:
            create_validator(model, config, results_dir, tensorboardx_logger, validation_data_loader, trainer)

        monitor = GpuMonitor(tensorboard_logger=tensorboardx_logger)

        results_root_dir = Path(config["general"]["results_dir"]).expanduser().resolve()
        mlflow.set_tracking_uri("file://" + str(results_root_dir / "mlflow"))

        # Get experiment_name and cast to string (it's a tomlkit.string by default)
        experiment_name = str(config["train"]["experiment_name"])

        # This will create the experiment if it doesn't exist
        mlflow.set_experiment(experiment_name)

        # If run_name is not `false` in the config, use it as the MLFlow run name in
        # this experiment. Otherwise use the name of the results directory
        run_name = str(config["train"]["run_name"]) if config["train"]["run_name"] else results_dir.name

        with mlflow.start_run(log_system_metrics=True, run_name=run_name):
            Train._log_params(config, results_dir)

            # Run the training process
            trainer.run(train_data_loader, max_epochs=config["train"]["epochs"])

        # Save the trained model
        model.save(results_dir / config["train"]["weights_filename"])
        with open(results_dir / "to_tensor.py", "w") as f:
            f.write(inspect.getsource(model.to_tensor))
        monitor.stop()

        logger.info("Finished Training")
        tensorboardx_logger.close()

        context = {
            "ml_framework": "pytorch",
            "results_dir": results_dir,
        }

        # Get a sample of input data. If the data is labeled, only return the input data.
        batch_sample = next(iter(train_data_loader))
        if isinstance(batch_sample, dict):
            batch_sample = model.to_tensor(batch_sample)
        sample = batch_sample[0] if isinstance(batch_sample, (list, tuple)) else batch_sample

        export_to_onnx(model, sample, config, context)

        return model

    @staticmethod
    def _log_params(config, results_dir):
        """Log the various parameters to mlflow from the config file.

        Parameters
        ----------
        config : dict
            The main configuration dictionary

        results_dir: str
            The full path to the results sub-directory
        """

        # Log full path to results subdirectory
        mlflow.log_param("Results Directory", results_dir)

        # Log all model params
        mlflow.log_params(config["model"])

        # Log some training and data loader params
        mlflow.log_param("epochs", config["train"]["epochs"])
        mlflow.log_param("batch_size", config["data_loader"]["batch_size"])

        # Log the criterion and optimizer params
        criterion_name = config["criterion"]["name"]
        mlflow.log_param("criterion", criterion_name)
        if criterion_name in config:
            mlflow.log_params(config[criterion_name])

        optimizer_name = config["optimizer"]["name"]
        mlflow.log_param("optimizer", optimizer_name)
        if optimizer_name in config:
            mlflow.log_params(config[optimizer_name])



import logging
from typing import Any

import numpy as np

from hyrax.data_sets.data_set_registry import DATA_SET_REGISTRY

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def generate_data_request_from_config(config):
    """This function handles the backward compatibility issue of defining the requested
    dataset in the `[data_set]` table in the config. If a `[model_inputs]` table
    is not defined, we will assemble a `data_request` dictionary from the values
    defined elsewhere in the configuration file.

    NOTE: We should anticipate deprecating the ability to define a data_request in
    `[data_set]`, when that happens, we should be able to remove this function.

    Parameters
    ----------
    config : dict
        The Hyrax configuration that can is passed to each dataset instance.

    Returns
    -------
    dict
        A dictionary where keys are dataset names and values are lists of fields
    """

    if "model_inputs" in config:
        data_request = config["model_inputs"]
    else:
        data_request = {
            "data": {
                "dataset_class": config["data_set"]["name"],
                "data_location": config["general"]["data_dir"],
                "primary_id_field": "object_id",
            },
        }

    return data_request


class DataProvider:
    """This class presents itself as a PyTorch Dataset, but acts like a GraphQL
    gateway that fetches data from multiple datasets based on the `model_inputs`
    dictionary provided during initialization.

    This class allows for flexible data retrieval from multiple dataset classes,
    each of which can have different fields requested.

    Additionally, the user can provide specific configuration options for each
    dataset class that will be merged with the original configuration provided
    during initialization.
    """

    def __init__(self, config: dict):
        """Initialize the DataProvider with a Hyrax config and extract (or create)
        the data_request.

        Parameters
        ----------
        config : dict
            The Hyrax configuration that defines the data_request.
        """

        self.config = config
        self.data_request = generate_data_request_from_config(self.config)

        self.validate_request(self.data_request)

        self.prepped_datasets = {}
        self.dataset_getters = {}
        self.all_metadata_fields = {}

        self.primary_dataset = None
        self.primary_dataset_id_field_name = None

        self.prepare_datasets()

        self.pull_up_primary_dataset_methods()

    def pull_up_primary_dataset_methods(self):
        """If a primary dataset is defined, we will pull up some of its methods
        to the DataProvider level so that they can be called directly on the
        DataProvider instance."""

        if self.primary_dataset:
            primary_dataset_instance = self.prepped_datasets[self.primary_dataset]

            # extend this tuple with more prefixes as needed
            exclude_prefixes = ("_", "get_")
            lifted_methods = [
                name
                for name in dir(primary_dataset_instance)
                if not any(name.startswith(p) for p in exclude_prefixes)
                and callable(getattr(primary_dataset_instance, name, None))
            ]

            for method_name in lifted_methods:
                if not hasattr(self, method_name):
                    setattr(self, method_name, getattr(primary_dataset_instance, method_name))

    def __getitem__(self, idx) -> dict:
        """This method returns data for a given index.

        It is also a wrapper that allows this class to be treated as a PyTorch
        Dataset.

        Parameters
        ----------
        idx : int
            The index of the data item to retrieve.

        Returns
        -------
        dict
            A dictionary containing the requested data from the prepared datasets.
        """
        return self.resolve_data(idx)

    def __len__(self) -> int:
        """Returns the length of the dataset.
        If the primary dataset is defined, it will return that length, otherwise
        it will use the length of the first dataset in ``self.prepped_datasets``.
        """
        return len(self._primary_or_first_dataset())

    def __repr__(self) -> str:
        repr_str = ""
        for friendly_name, data in self.data_request.items():
            if isinstance(data, dict):
                repr_str += f"Name: {friendly_name}\n"
                repr_str += f"  Dataset class: {data['dataset_class']}\n"
                if "data_location" in data:
                    repr_str += f"  Data location: {data['data_location']}\n"
                if "fields" in data:
                    repr_str += f"  Requested fields: {', '.join(data.get('fields', []))}\n"
                else:
                    repr_str += "  Requested fields: *All available fields*\n"
                if "dataset_config" in data:
                    repr_str += "  Dataset config:\n"
                    for k, v in data["dataset_config"].items():
                        repr_str += f"    {k}: {v}\n"
        return repr_str

    def fields(self) -> dict:
        """Print all the available fields for each dataset in the DataProvider.

        Returns
        -------
        dict
            A dictionary mapping friendly dataset names to their available fields.
        """
        fields_dict: dict[str, list[str]] = {}
        for friendly_name, fields in self.dataset_getters.items():
            fields_dict[friendly_name] = list(fields.keys())
        return fields_dict

    def is_iterable(self):
        """DataProvider datasets will always be map-style datasets."""
        return False

    def is_map(self):
        """DataProvider datasets will always be map-style datasets."""
        return True

    # ^ Since the getter methods are dynamically created when a dataset class is _instantiated_,
    # ^ we can't really validate that the requested fields exist until after instantiation.
    # ^ And instantiation doesn't happen here, it happens in `prepare_datasets`.
    # ^ See: https://github.com/lincc-frameworks/hyrax/issues/419

    @staticmethod
    def validate_request(data_request: dict):
        """Convenience method to ensure that each requested dataset exists and that
        each field in each dataset has a `get_<field_name>` method."""
        problem_count = 0
        for friendly_name, dataset_parameters in data_request.items():
            dataset_class = dataset_parameters.get("dataset_class")
            if not dataset_class:
                logger.error(f"Model input for '{friendly_name}' does not specify a 'dataset_class'.")
                problem_count += 1
                continue
            if dataset_class not in DATA_SET_REGISTRY:
                logger.error(
                    f"Unable to locate dataset, '{dataset_class}' in the registered datasets:"
                    f" {list(DATA_SET_REGISTRY.keys())}."
                )
                problem_count += 1
                continue
            if DATA_SET_REGISTRY[dataset_class].is_iterable():
                logger.error(
                    f"Dataset '{dataset_class}' is an iterable-style dataset. "
                    "This is not supported in the current implementation of DataProvider. "
                    "Hyrax DataProvider only supports map-style datasets at this time. "
                    "You should instantiate an iterable-style dataset class directly."
                )
                problem_count += 1

        if problem_count > 0:
            logger.error(f"Finished validating request. Problems found: {problem_count}")
            raise RuntimeError("Data request validation failed. See logs for details.")

    def prepare_datasets(self):
        """Instantiate each of the requested datasets based on the ``model_inputs``
        configuration dictionary. Store the prepared instances in the
        ``self.prepped_datasets`` dictionary."""

        # Note: We can be less strict about checking for existence of keys here
        # because we have already validated the ``model_inputs`` in
        # `self.validate_request()`.
        for friendly_name, dataset_definition in self.data_request.items():
            dataset_class = dataset_definition.get("dataset_class")
            data_location = dataset_definition.get("data_location")

            # Create a temporary config dictionary that merges the original
            # config with the dataset-specific config.
            dataset_specific_config = self._apply_configurations(self.config, dataset_definition)

            # Instantiate the dataset class
            dataset_instance = DATA_SET_REGISTRY[dataset_class](
                config=dataset_specific_config, data_location=data_location
            )

            # Store the prepared dataset instance in the `self.prepped_datasets`
            self.prepped_datasets[friendly_name] = dataset_instance

            # If no fields were specifically requested, we'll assume that the user
            # wants _all_ the available fields - user defined and dynamically created!
            if not dataset_definition.get("fields", []):
                dataset_definition["fields"] = [
                    method[4:] for method in dir(dataset_instance) if method.startswith("get_")
                ]

            for field in dataset_definition.get("fields", []):
                if not hasattr(dataset_instance, f"get_{field}"):
                    logger.error(
                        f"No `get_{field}` method for requested field, '{field}' "
                        f"was found in dataset {dataset_class}."
                    )

            # Cache all of the `get_<field_name>` methods in the dataset instance
            # so that we don't have to look them up each time we call `resolve_data`.
            self.dataset_getters[friendly_name] = {}
            for method in dir(dataset_instance):
                if method.startswith("get_"):
                    field_name = method[4:]  # Remove the "get_" prefix
                    self.dataset_getters[friendly_name][field_name] = getattr(dataset_instance, method)

            if len(self.dataset_getters[friendly_name]) == 0:
                logger.error(
                    f"No `get_*` methods were found in the class: {dataset_class}. "
                    "This is likely an error in the dataset class definition."
                )

            # Get all the dataset's metadata fields and store them in
            # `self.all_metadata_fields` dictionary. Modify the name to be
            # <metadata_field_name>_<friendly_name>, i.e. "RA_cifar" or "photoz_hsc".
            if dataset_instance._metadata_table:
                columns = [f"{col}_{friendly_name}" for col in dataset_instance._metadata_table.colnames]
                self.all_metadata_fields[friendly_name] = columns
            else:
                self.all_metadata_fields[friendly_name] = []

            # If this dataset is marked as the primary dataset, store that
            # information for later use.
            if "primary_id_field" in dataset_definition:
                self.primary_dataset = friendly_name
                self.primary_dataset_id_field_name = dataset_definition["primary_id_field"]

    @staticmethod
    def _apply_configurations(base_config: dict, dataset_definition: dict) -> dict:
        """Merge the original base config with the dataset-specific config.

        This function uses ``ConfigManager.merge_configs`` to merge the
        dataset-specific configuration into a copy of the original base config.

        If no ``dataset_config`` is provided in the ``dataset_definition`` dict,
        the original base config will be returned unmodified.

        Example of a dataset definition dictionary:

        .. code-block:: python

            "my_dataset": {
                "dataset_class": "MyDataset",
                "data_location": "/path/to/data",
                "dataset_config": {
                    "param1": "value1",
                    "param2": "value2"
                },
                "fields": ["field1", "field2"]
            }

        or equivalently in a .toml file:

        .. code-block:: toml

            [model_inputs.my_dataset]
            dataset_class = "MyDataset"
            data_location = "/path/to/data"
            fields = ["field1", "field2"]
            [model_inputs.my_dataset.dataset_config]
            param1 = "value1"
            param2 = "value2"

        In this example, the ``dataset_config`` dictionary will be merged into
        the original base config, overriding the values of param1 and param2
        when creating an instance of ``MyDataset``.

        Parameters
        ----------
        base_config : dict
            The original base configuration dictionary. A copy of this is created,
            the dataset_definition dict is merged into the copy, and the copy
            is returned.

        dataset_definition : dict
            A dictionary defining the dataset, including any dataset-specific
            configuration options in a nested ``dataset_config`` dictionary.

        Returns
        -------
        dict
            A final configuration dictionary to be passed when creating an instance
            of the dataset class.
        """
        from hyrax.config_utils import ConfigManager

        cm = ConfigManager()

        # ^ NOTE: This assumes that the dataset-specific configuration options
        # ^ are nested under a top-level key that matches the dataset class name.
        # ^ i.e. "data_set": {"MyDataset": {<dataset-specific-options>}}. Or in toml
        # ^ [data_set.MyDataset]
        # ^ <dataset-specific-options>
        # ^ See: https://github.com/lincc-frameworks/hyrax/issues/417

        if "dataset_config" in dataset_definition:
            tmp_config = {
                "data_set": {dataset_definition["dataset_class"]: dataset_definition["dataset_config"]}
            }

            # Note that `merge_configs` makes a copy of self.config, so the original
            # config will not be modified.
            return cm.merge_configs(base_config, tmp_config)
        else:
            return base_config

    def sample_data(self) -> dict:
        """Returns a data sample. Primarily this will be used for instantiating a
        model so that any runtime resizing can be handled properly.

        Returns
        -------
        dict
            A dictionary containing the data for index 0.
        """
        return self[0]

    # ^ What is the appropriate return when there is no ``ids()`` method in the
    # ^ primary_or_first dataset? Perhaps a generator that yields stop iteration error?
    def ids(self):
        """Returns the IDs of the dataset.

        If the primary dataset is defined it will return those ids, if not,
        it will return the ids of the first dataset in the list of
        prepped_dataset.keys()."""

        primary_dataset = self._primary_or_first_dataset()
        return primary_dataset.ids() if hasattr(primary_dataset, "ids") else []

    def resolve_data(self, idx: int) -> dict:
        """This method requests the field data from the prepared datasets by index.

        Parameters
        ----------
        idx : int
            The index of the data item to retrieve.

        Returns
        -------
        dict
            A dictionary containing the requested data from the prepared datasets.
        """
        returned_data: dict[str, dict[str, Any]] = {}
        for friendly_name in self.data_request:
            returned_data[friendly_name] = {}
            dataset_definition = self.data_request.get(friendly_name)

            # For each of the requested fields, call the corresponding
            # `get_<field_name>` method in the dataset instance.
            for field in dataset_definition.get("fields", []):
                returned_data[friendly_name][field] = self.dataset_getters[friendly_name][field](idx)

        # Because there is machinery in the consuming code that expects an "object_id"
        # key in the returned data, we will add that here if a primary dataset.
        if self.primary_dataset:
            returned_data["object_id"] = returned_data[self.primary_dataset][
                self.primary_dataset_id_field_name
            ]

        return returned_data

    # ^ If we move toward supporting get_<metadata_column_name> methods in datasets,
    # ^ we should be able to remove most or all of this method and the metadata_fields method.
    # ^ This is really here to support the visualization code, and if we convert that
    # ^ to using get_<metadata_column_name> methods, we can remove this.
    # ^ See: https://github.com/lincc-frameworks/hyrax/issues/418

    def metadata(self, idxs=None, fields=None) -> np.ndarray:
        """Fetch the requested metadata fields for the given indices.

        Example:

        .. code-block:: python

            # Fetch the metadata_1 and metadata_2 fields from the dataset with the
            # friendly name "random_1".

            metadata = data_provider.metadata(
                idxs=[0, 1, 2],
                fields=["metadata_1_random_1", "metadata_2_random_1"]
            )

        Parameters
        ----------
        idxs : list of int, optional
            A list of indices for which to fetch metadata. If None, no metadata
            will be returned.
        fields : list of str, optional
            A list of metadata fields to fetch. If None, no metadata will be
            returned.

        Returns
        -------
        np.ndarray
            A structured NumPy array containing the requested metadata fields.
            The dtype names of the array will be the metadata field names, modified
            to include the friendly name of the dataset they come from. For example,
            if the "RA" field comes from a dataset with the friendly name "cifar",
            the returned field name will be "RA_cifar".
        """

        if idxs is None:
            idxs = []

        if fields is None:
            fields = []

        # Create an empty structured array to hold the merged metadata
        returned_metadata = np.empty(0, dtype=[])

        # For each dataset:
        # 1) Find the requested metadata fields that come from it
        # 2) Strip the friendly name from the metadata field name
        # 3) Call the dataset's `metadata` method with indices and metadata fields.
        for friendly_name, dataset in self.prepped_datasets.items():
            metadata_fields_to_fetch = [
                field.replace(f"_{friendly_name}", "")
                for field in fields
                if field.endswith(f"_{friendly_name}")
            ]

            if metadata_fields_to_fetch:
                this_metadata = dataset.metadata(idxs, metadata_fields_to_fetch)
                # Append the friendly name to the columns
                this_metadata.dtype.names = [f"{name}_{friendly_name}" for name in this_metadata.dtype.names]

                # merge this_metadata into the returned_metadata structured array
                if returned_metadata.size == 0:
                    returned_metadata = this_metadata
                else:
                    returned_metadata = np.lib.recfunctions.merge_arrays(
                        (returned_metadata, this_metadata), flatten=True
                    )

        return returned_metadata

    def metadata_fields(self, friendly_name=None) -> list[str]:
        """Returns a list of metadata fields that are available across all prepared
        datasets.

        The field names will be modified to include the friendly name of the
        dataset they come from. For example, if the "RA" field comes from a dataset
        with the friendly name "cifar", the returned field name will be "RA_cifar".

        NOTE: If a specific dataset friendly_name is provided, only the metadata
        fields for that dataset will be returned, and the field names will not
        include the friendly name suffix.

        Parameters
        ----------
        friendly_name : str, optional
            If provided, only the metadata fields for the specified friendly name
            will be returned. If not provided, metadata fields from all datasets
            will be returned.

        Returns
        -------
        list[str]
            The column names of the metadata table passed. Empty list if no metadata
            was provided during construction of the DataProvider.
        """
        all_fields = []
        if friendly_name:
            return [
                field.replace(f"_{friendly_name}", "")
                for field in self.all_metadata_fields.get(friendly_name, [])
            ]

        for _, v in self.all_metadata_fields.items():
            all_fields.extend(v)

        # Always include the `object_id` field
        all_fields.append("object_id")

        return all_fields

    def _primary_or_first_dataset(self):
        """Returns the primary dataset instance if it exists, otherwise returns
        the first dataset in the prepped_datasets."""

        # Get the list of friendly names for the prepared datasets
        keys = list(self.prepped_datasets.keys())

        # If a primary dataset is defined, use that, otherwise use the first one
        dataset_to_use = self.primary_dataset if self.primary_dataset else keys[0]

        return self.prepped_datasets[dataset_to_use]



import builtins
from unittest.mock import patch

import pytest

from hyrax import Hyrax
from hyrax.data_sets.data_provider import DataProvider, generate_data_request_from_config


def test_generate_data_request_from_config():
    """Test that we support generating a data request dictionary
    outside of the `model_inputs` table."""

    h = Hyrax()
    config = dict(h.config)
    config.pop("model_inputs", None)

    config["data_set"]["name"] = "HyraxRandomDataset"
    config["general"]["data_dir"] = "./data"

    ret_val = generate_data_request_from_config(config)

    assert "data" in ret_val
    assert "dataset_class" in ret_val["data"]
    assert ret_val["data"]["dataset_class"] == "HyraxRandomDataset"
    assert "data_location" in ret_val["data"]
    assert ret_val["data"]["data_location"] == "./data"


def test_generate_data_request_passes_model_inputs():
    """Test that generate_data_request passes the model_inputs
    dict from the config, unchanged."""

    h = Hyrax()
    h.config["model_inputs"] = {
        "a": "foo",
        "b": {"c": "bar"},
    }

    ret_val = generate_data_request_from_config(h.config)

    assert "a" in ret_val
    assert ret_val["a"] == "foo"
    assert "b" in ret_val
    assert isinstance(ret_val["b"], dict)
    assert ret_val["b"] == {"c": "bar"}


def test_data_provider(data_provider):
    """Testing the happy path scenario of creating a DataProvider
    instance with a config that requests two instances of
    `HyraxRandomDataset`.
    """

    dp = data_provider

    assert dp.primary_dataset == "random_0"
    assert dp.primary_dataset_id_field_name == "object_id"

    assert dp.is_iterable() is False
    assert dp.is_map() is True

    # There should be 2 prepared datasets
    assert len(dp.prepped_datasets) == 2
    assert "random_0" in dp.prepped_datasets
    assert "random_1" in dp.prepped_datasets

    # There should be 2 dataset_getters dicts with subdicts of different sizes
    assert len(dp.dataset_getters) == 2
    assert len(dp.dataset_getters["random_0"]) == 5
    assert len(dp.dataset_getters["random_1"]) == 5

    data_request = dp.data_request
    for friendly_name in data_request:
        for field in data_request[friendly_name]["fields"]:
            assert field in dp.dataset_getters[friendly_name]

    data_request = dp.data_request
    for friendly_name in data_request:
        assert len(dp.all_metadata_fields[friendly_name]) == 3
        for metadata_field in dp.all_metadata_fields[friendly_name]:
            assert friendly_name in metadata_field


def test_validate_request(multimodal_config):
    """Basic test to see that validation works as expected in the base
    case."""

    c = multimodal_config
    DataProvider.validate_request(c)


def test_validate_request_no_dataset_class(multimodal_config, caplog):
    """Basic test to see that validation works as when no dataset class
    name is provided."""

    c = multimodal_config
    c["random_0"].pop("dataset_class", None)
    with caplog.at_level("ERROR"):
        with pytest.raises(RuntimeError) as execinfo:
            DataProvider.validate_request(c)

    assert "failed" in str(execinfo.value)
    assert "does not specify a 'dataset_class'" in caplog.text


def test_validate_request_unknown_dataset(multimodal_config, caplog):
    """Basic test to see that validation raises correctly when a nonexistent
    dataset class name is provided."""

    c = multimodal_config
    c["random_0"]["dataset_class"] = "NoSuchDataset"
    with caplog.at_level("ERROR"):
        with pytest.raises(RuntimeError) as execinfo:
            DataProvider.validate_request(c)

    assert "failed" in str(execinfo.value)
    assert "Unable to locate dataset" in caplog.text


def test_validate_request_iterable_dataset(multimodal_config, caplog):
    """Basic test to see that validation works correctly when an iterable dataset
    is requested"""

    c = multimodal_config
    c["random_0"]["dataset_class"] = "HyraxRandomIterableDataset"
    with caplog.at_level("ERROR"):
        with pytest.raises(RuntimeError) as execinfo:
            DataProvider.validate_request(c)

    assert "failed" in str(execinfo.value)
    assert "is an iterable-style dataset" in caplog.text


def test_validate_request_bad_field(multimodal_config, caplog):
    """Basic test to see that validation works correctly when a bad field is
    requested."""
    h = Hyrax()
    c = multimodal_config
    c["random_0"]["fields"] = ["image", "no_such_field"]
    h.config["model_inputs"] = c
    with caplog.at_level("ERROR"):
        DataProvider(h.config)

    assert "No `get_no_such_field` method" in caplog.text


def test_validate_request_dataset_missing_getters(multimodal_config, caplog):
    """Basic test to see that validation works correctly when a dataset is
    missing all getters."""

    h = Hyrax()
    c = multimodal_config
    c["random_0"].pop("fields", None)
    h.config["model_inputs"] = c

    # Fake methods to return from `dir`, none of which start with `get_*`.
    fake_methods = ["fake_one", "fake_two", "fake_three"]

    with patch.object(builtins, "dir", return_value=fake_methods):
        with caplog.at_level("ERROR"):
            DataProvider(h.config)

    assert "No `get_*` methods were found" in caplog.text


def test_apply_configurations(multimodal_config):
    """Test the static method _apply_configurations to ensure that
    it merges a base config with a dataset-specific config correctly."""

    from hyrax import Hyrax

    h = Hyrax()
    base_config = h.config
    model_inputs = multimodal_config

    merged_config = DataProvider._apply_configurations(base_config, model_inputs["random_0"])

    assert merged_config["data_set"]["HyraxRandomDataset"]["shape"] == [2, 16, 16]
    assert (
        base_config["data_set"]["HyraxRandomDataset"]["seed"]
        == merged_config["data_set"]["HyraxRandomDataset"]["seed"]
    )
    assert merged_config["general"] == base_config["general"]

    merged_config = DataProvider._apply_configurations(base_config, model_inputs["random_1"])

    assert base_config["data_set"]["HyraxRandomDataset"]["shape"] != [5, 16, 16]
    assert base_config["data_set"]["HyraxRandomDataset"]["seed"] != 4200
    assert merged_config["data_set"]["HyraxRandomDataset"]["shape"] == [5, 16, 16]
    assert merged_config["data_set"]["HyraxRandomDataset"]["seed"] == 4200
    assert merged_config["general"] == base_config["general"]


def test_primary_or_first_dataset(multimodal_config):
    """Test that if no primary dataset is specified, the first dataset
    in the config is returned."""

    from hyrax import Hyrax

    h = Hyrax()

    # Base case with `primary_id_field` defined on `random_0`
    model_inputs = multimodal_config
    h.config["model_inputs"] = model_inputs

    dp = DataProvider(h.config)
    dp.prepare_datasets()

    assert dp.primary_dataset == "random_0"
    assert dp.primary_dataset_id_field_name == "object_id"

    primary_dataset = dp._primary_or_first_dataset()
    assert primary_dataset.data_location == "./in_memory_0"

    # Secondary case with no `primary_id_field` defined
    model_inputs["random_0"].pop("primary_id_field", None)
    h.config["model_inputs"] = model_inputs

    dp = DataProvider(h.config)
    dp.prepare_datasets()

    assert dp.primary_dataset is None
    assert dp.primary_dataset_id_field_name is None

    primary_dataset = dp._primary_or_first_dataset()
    assert primary_dataset.data_location == "./in_memory_0"

    # Tertiary case with `primary_id_field` defined on `random_1`
    model_inputs["random_1"]["primary_id_field"] = "object_id"
    h.config["model_inputs"] = model_inputs

    dp = DataProvider(h.config)
    dp.prepare_datasets()

    assert dp.primary_dataset == "random_1"
    assert dp.primary_dataset_id_field_name == "object_id"

    primary_dataset = dp._primary_or_first_dataset()
    assert primary_dataset.data_location == "./in_memory_1"


def test_metadata_fields(data_provider):
    """Test that the calling metadata_fields returns the expected
    fields with the expected structure."""

    dp = data_provider
    dp.prepare_datasets()

    all_metadata_fields = dp.all_metadata_fields

    assert "random_0" in all_metadata_fields
    assert "random_1" in all_metadata_fields

    assert len(all_metadata_fields["random_0"]) == 3
    assert len(all_metadata_fields["random_1"]) == 3

    all_fields = dp.metadata_fields()

    assert isinstance(all_fields, list)
    assert "object_id" in all_fields

    expected_metadata_fields = ["object_id", "meta_field_1", "meta_field_2"]
    for field in expected_metadata_fields:
        assert field + "_random_0" in all_fields
        assert field + "_random_1" in all_fields


def test_metadata_fields_with_friendly_name(data_provider):
    """Test that the calling metadata_fields returns the expected
    fields with the expected structure."""

    dp = data_provider
    dp.prepare_datasets()

    all_fields = dp.metadata_fields("random_0")

    assert isinstance(all_fields, list)
    assert "object_id" in all_fields

    expected_metadata_fields = ["object_id", "meta_field_1", "meta_field_2"]
    for field in expected_metadata_fields:
        assert field in all_fields


def test_sample_data():
    """Test that sample_data returns a dictionary with the expected
    structure.

    We don't use the test fixture here so that this can be a little more
    flexible and self-contained.


    The expected result structure is:
    {
        'random_0': {
            'object_id': <int>,
            'image': array(...),
            'label': 'cat'
        },
        'random_1': {
            'image': array(...)
        },
        'object_id': <int>
    }
    """
    from hyrax import Hyrax

    h = Hyrax()

    multimodal_config = {
        "random_0": {
            "dataset_class": "HyraxRandomDataset",
            "data_directory": "./in_memory_0",
            "fields": ["object_id", "image", "label"],
            "dataset_config": {
                "shape": [2, 16, 16],
            },
            "primary_id_field": "object_id",
        },
        "random_1": {
            "dataset_class": "HyraxRandomDataset",
            "data_directory": "./in_memory_1",
            "fields": ["image"],
            "dataset_config": {
                "shape": [5, 16, 16],
                "seed": 4200,
            },
        },
    }

    h.config["model_inputs"] = multimodal_config
    dp = DataProvider(h.config)
    dp.prepare_datasets()

    sample = dp.sample_data()

    assert isinstance(sample, dict)
    assert "random_0" in sample
    assert "random_1" in sample
    assert "object_id" in sample
    assert len(sample) == 3
    assert isinstance(sample["random_0"], dict)
    assert isinstance(sample["random_1"], dict)
    assert len(sample["random_0"]) == 3
    assert len(sample["random_1"]) == 1

    for friendly_name in ["random_0", "random_1"]:
        dataset_sample = sample[friendly_name]
        assert "image" in dataset_sample

        if friendly_name == "random_0":
            assert "object_id" in dataset_sample
            assert "label" in dataset_sample


def test_data_provider_get_item(data_provider):
    """Basic test to ensure that different index values return different data."""
    dp = data_provider
    dp.prepare_datasets()

    sample_0a = dp[0]
    sample_0b = dp.resolve_data(0)
    sample_1a = dp[1]
    sample_1b = dp.resolve_data(1)

    assert isinstance(sample_0a, dict)
    assert isinstance(sample_1a, dict)
    assert sample_0a["random_0"]["image"][0][0][0] != sample_1a["random_0"]["image"][0][0][0]
    assert sample_0a["random_0"]["image"][0][0][0] == sample_0b["random_0"]["image"][0][0][0]

    assert isinstance(sample_0b, dict)
    assert isinstance(sample_1b, dict)
    assert sample_0b["random_0"]["image"][0][0][0] != sample_1b["random_0"]["image"][0][0][0]
    assert sample_1a["random_0"]["image"][0][0][0] == sample_1b["random_0"]["image"][0][0][0]

    assert sample_0a["random_0"]["object_id"] == sample_0b["random_0"]["object_id"]
    assert sample_1a["random_0"]["object_id"] == sample_1b["random_0"]["object_id"]
    assert "object_id" in sample_0a
    assert "object_id" in sample_1a
    assert "object_id" in sample_0b
    assert "object_id" in sample_1b


def test_data_provider_returns_length(data_provider):
    """Basic test to ensure that __len__ returns the expected value.
    The length returned from DataProvider is the length of the primary
    dataset or, if no primary dataset is specified.

    We have specified random_0 to be the primary dataset in conftest.
    """

    dp = data_provider
    dp.prepare_datasets()

    length = len(dp)

    random_0_length = len(dp.prepped_datasets["random_0"])
    assert isinstance(length, int)
    assert length == random_0_length


def test_data_provider_ids(data_provider):
    """Basic test to ensure that ids() returns the expected value.
    The ids returned from DataProvider are the ids of the primary
    dataset or, if no primary dataset is specified, the first dataset.

    We have specified random_0 to be the primary dataset in conftest.
    """

    dp = data_provider
    dp.prepare_datasets()

    ids = list(dp.ids())

    random_0_ids = list(dp.prepped_datasets["random_0"].ids())
    random_1_ids = list(dp.prepped_datasets["random_1"].ids())

    assert len(ids) == len(random_0_ids)
    assert all(i == j for i, j in zip(ids, random_0_ids))
    assert all(i != j for i, j in zip(ids, random_1_ids))


def test_data_provider_returns_metadata(data_provider):
    """Basic test to ensure that metadata() returns the expected value.
    The metadata returned from DataProvider are the metadata of the primary
    dataset or, if no primary dataset is specified, the first dataset.

    We have specified random_0 to be the primary dataset in conftest.
    """

    dp = data_provider
    dp.prepare_datasets()

    metadata = dp.metadata()
    assert len(metadata) == 0

    metadata = dp.metadata(idxs=None, fields=[])
    assert len(metadata) == 0

    metadata = dp.metadata(idxs=[], fields=[])
    assert len(metadata) == 0

    metadata = dp.metadata(idxs=[], fields=["meta_field_1_random_0"])
    assert len(metadata) == 0
    assert len(metadata.dtype.names) == 1
    assert "meta_field_1_random_0" in metadata.dtype.names[0]

    metadata = dp.metadata(idxs=[5], fields=["meta_field_1_random_0"])
    assert len(metadata) == 1
    assert len(metadata.dtype.names) == 1
    assert "meta_field_1_random_0" in metadata.dtype.names[0]

    metadata = dp.metadata(idxs=[5, 97], fields=["meta_field_1_random_0"])
    assert len(metadata) == 2
    assert len(metadata.dtype.names) == 1
    assert "meta_field_1_random_0" in metadata.dtype.names

    metadata = dp.metadata(idxs=[5, 97], fields=["meta_field_1_random_0", "meta_field_2_random_1"])
    assert len(metadata) == 2
    assert len(metadata.dtype.names) == 2
    assert "meta_field_1_random_0" in metadata.dtype.names
    assert "meta_field_2_random_1" in metadata.dtype.names



import logging
import sys

import numpy as np
import pytest
from astropy.table import Table
from torch import from_numpy
from torch.utils.data import Dataset, IterableDataset

import hyrax
import hyrax.data_sets
from hyrax.data_sets import HyraxDataset
from hyrax.data_sets.data_provider import DataProvider

logger = logging.getLogger(__name__)


def pytest_configure(config):
    """
    Global test configuration. We:
    1) Disable ConfigManager from slurping up files from the working directory to enable test reproducibility
       across different developer machines and CI.

    2) Set an unlimited number of open files per process on OSX. OSX's default per-process file limit is 256
       Because we use temporary files during many of our tests, it's easy to go over this limit.
    """
    hyrax.config_utils.ConfigManager._called_from_test = True

    if sys.platform == "darwin":
        import resource

        try:
            resource.setrlimit(resource.RLIMIT_NOFILE, (resource.RLIM_INFINITY, resource.RLIM_INFINITY))
        except ValueError as e:
            msg = "Attempted to raise open file limit, and failed. Tests may not work.\n"
            msg += f"See error below when trying to raise open file limit: \n {e}"
            raise RuntimeError(msg) from e


class RandomDataset(HyraxDataset, Dataset):
    """Dataset yielding pairs of random numbers. Requires a seed to emulate
    static data on the filesystem between instantiations"""

    def __init__(self, config, data_directory):
        size = config["data_set"]["size"]

        dim_1_length = 2
        if "dimension_1_length" in config["data_set"]:
            dim_1_length = config["data_set"]["dimension_1_length"]

        dim_2_length = 0
        if "dimension_2_length" in config["data_set"]:
            dim_2_length = config["data_set"]["dimension_2_length"]

        seed = config["data_set"]["seed"]
        rng = np.random.default_rng(seed)

        print(f"Initialized dataset with dim 1: {dim_1_length}, dim 2: {dim_2_length}")

        if dim_2_length > 0:
            self.data = rng.random((size, dim_1_length, dim_2_length), np.float32)
        else:
            self.data = rng.random((size, dim_1_length), np.float32)

        # Start our IDs at a random integer between 0 and 100
        id_start = rng.integers(100)
        self.id_list = list(range(id_start, id_start + size))

        metadata_table = Table({"object_id": np.array(list(self.ids()))})

        super().__init__(config, metadata_table)

    def __getitem__(self, idx):
        return from_numpy(self.data[idx])

    def __len__(self):
        return len(self.data)

    def ids(self):
        """Yield IDs for the dataset"""
        for id_item in self.id_list:
            yield str(id_item)

    def get_ids(self, idx):
        """Returns the ids given an index."""
        return self.id_list[idx]


class RandomIterableDataset(RandomDataset, IterableDataset):
    """Iterable version of RandomDataset"""

    def __iter__(self):
        for item in self.data:
            yield from_numpy(item)


@pytest.fixture(scope="function", params=["HyraxRandomDataset", "HyraxRandomIterableDataset"])
def loopback_hyrax(tmp_path_factory, request):
    """This generates a loopback hyrax instance
    which is configured to use the loopback model
    and a simple dataset yielding random numbers
    """
    results_dir = tmp_path_factory.mktemp(f"loopback_hyrax_{request.param}")

    h = hyrax.Hyrax()
    h.config["model"]["name"] = "HyraxLoopback"
    h.config["train"]["epochs"] = 1
    h.config["data_loader"]["batch_size"] = 5
    h.config["general"]["results_dir"] = str(results_dir)

    h.config["general"]["dev_mode"] = True
    h.config["data_set"]["name"] = request.param
    h.config["data_set"]["HyraxRandomDataset"]["size"] = 20
    h.config["data_set"]["HyraxRandomDataset"]["seed"] = 0
    h.config["data_set"]["HyraxRandomDataset"]["shape"] = [2, 3]

    h.config["data_set"]["validate_size"] = 0.2
    h.config["data_set"]["test_size"] = 0.2
    h.config["data_set"]["train_size"] = 0.6

    weights_file = results_dir / "fakeweights"
    with open(weights_file, "a"):
        pass
    h.config["infer"]["model_weights_file"] = str(weights_file)

    dataset = h.prepare()
    return h, dataset


@pytest.fixture(scope="function")
def loopback_inferred_hyrax(loopback_hyrax):
    """This generates a loopback hyrax instance which is configured to use the
    loopback model and a simple dataset yielding random numbers. It includes a call
    to hyrax.infer which will produce the output consumed by vdb_index or umap."""

    h, dataset = loopback_hyrax
    inference_results = h.infer()

    return h, dataset, inference_results


@pytest.fixture(scope="function")
def multimodal_config():
    """Create a hyrax instance with a default config setting, then update the
    config to represent a request for multimodal data."""

    return {
        "random_0": {
            "dataset_class": "HyraxRandomDataset",
            "data_location": "./in_memory_0",
            "fields": ["object_id", "image", "label"],
            "dataset_config": {
                "shape": [2, 16, 16],
            },
            "primary_id_field": "object_id",
        },
        "random_1": {
            "dataset_class": "HyraxRandomDataset",
            "data_location": "./in_memory_1",
            "fields": ["image"],
            "dataset_config": {
                "shape": [5, 16, 16],
                "seed": 4200,
            },
        },
    }


@pytest.fixture(scope="function")
def data_provider(multimodal_config):
    """Use the multimodal_config fixture to create a DataProvider instance."""
    h = hyrax.Hyrax()
    h.config["model_inputs"] = multimodal_config
    dp = DataProvider(h.config)
    return dp



# ruff: noqa: D101, D102
import logging
from pathlib import Path

import numpy as np
from torch.utils.data import Dataset, IterableDataset

from hyrax.config_utils import ConfigDict

from .data_set_registry import HyraxDataset

logger = logging.getLogger(__name__)


class HyraxCifarBase:
    """Base class for Hyrax Cifar datasets"""

    def __init__(self, config: ConfigDict, data_location: Path = None):
        import torchvision.transforms as transforms
        from astropy.table import Table
        from torchvision.datasets import CIFAR10

        self.data_location = data_location if data_location else config["general"]["data_dir"]

        transform = transforms.Compose(
            [transforms.ToTensor(), transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))]
        )
        self.cifar = CIFAR10(root=self.data_location, train=True, download=True, transform=transform)
        metadata_table = Table(
            {"label": np.array([self.cifar[index][1] for index in range(len(self.cifar))])}
        )
        super().__init__(config, metadata_table)

    def get_image(self, idx):
        """Get the image at the given index as a NumPy array."""
        image, _ = self.cifar[idx]
        return image.numpy()

    def get_label(self, idx):
        """Get the label at the given index."""
        _, label = self.cifar[idx]
        return label

    def get_index(self, idx):
        """Get the index of the item."""
        return idx

    def get_object_id(self, idx):
        """Get the object ID for the item."""
        return idx


class HyraxCifarDataSet(HyraxCifarBase, HyraxDataset, Dataset):
    """Map style CIFAR 10 dataset for Hyrax

    This is simply a version of CIFAR10 that is initialized using Hyrax config with a transformation
    that works well for example code.

    We only use the training split in the data, because it is larger (50k images). Hyrax will then divide that
    into Train/test/Validate according to configuration.
    """

    def __len__(self):
        return len(self.cifar)

    def __getitem__(self, idx):
        return {
            "data": {
                "object_id": self.get_object_id(idx),
                "image": self.get_image(idx),
                "label": self.get_label(idx),
            },
            "object_id": self.get_object_id(idx),
        }


class HyraxCifarIterableDataSet(HyraxCifarBase, HyraxDataset, IterableDataset):
    """Iterable style CIFAR 10 dataset for Hyrax

    This is simply a version of CIFAR10 that is initialized using Hyrax config with a transformation
    that works well for example code. This version only supports iteration, and not map-style access

    We only use the training split in the data, because it is larger (50k images). Hyrax will then divide that
    into Train/test/Validate according to configuration.
    """

    def __iter__(self):
        for idx in range(len(self.cifar)):
            yield {
                "data": {
                    "object_id": self.get_object_id(idx),
                    "image": self.get_image(idx),
                    "label": self.get_label(idx),
                },
                "object_id": self.get_object_id(idx),
            }
